// Dame.java                                        MM 2014
// Model

/**
 * Loest Damenproblem auf einem 8*8 Schachbrett
 * von C. F. Gauss, gibt alle 12 Loesungen
 * ohne Rotation und Spiegelung aus.
 * @author Monika Meiler
 */
public class Dame
{
/* ----------------------------------------------------- */
                                                // coreData
/**
 * Schachbrett,
 * jede Spaltenkomponente enthaelt Zeilennummer,
 * in der eine Dame steht, baut Loesung auf.
 */
  private int spalten[ ] = new int[ 8];

/**
 * Bedrohungen durch bereits gesetzte Dame in Zeilen.
 */
  private boolean zeilen[ ] = new boolean[ 8];

/**
 * Bedrohungen durch bereits gesetzte Dame in den 
 * positiven Diagonalen.
 */
  private boolean posDiag[ ] = new boolean[ 16];

/**
 * Bedrohungen durch bereits gesetzte Dame in den 
 * negativen Diagonalen.
 */
  private boolean negDiag[ ] = new boolean[ 16];

/**
 * Loesungsmenge.
 */
  private Menge<DameLoesung> loesungen;

/* ----------------------------------------------------- */
                                             // Konstruktor
/**
 * Konstruktor, setzt alle Bedrohungen auf false.
 */
  public Dame()
  {
    reset();
  }

/* ----------------------------------------------------- */
                                         // getData-Methode
/*
 * Gibt Dameloesungsmenge zurueck.
 */
  public Menge<DameLoesung> getLoesungen()
  {
    return loesungen;
  }

/* ----------------------------------------------------- */
                                        // Service-Methoden
/**
 * Stellt Startzustand her.
 */
  public void reset()
  {
    int i;
    for( i = 0; i < 8; i++) zeilen[ i] = false;
    for( i = 0; i < 16; i++)
      posDiag[ i] = negDiag[ i] = false;

    loesungen = new Menge<DameLoesung>();
  }

/* ----------------------------------------------------- */
                                               // Rekursion
/**
 * Startet die Rekursion.
 */
  public void start()
  {
    setzeDame( 0);         // Aufruf der rekursiven Methode
  }

/**
 * Suchen der Loesungen, rekursiv.
 * @param spalte Spalte, in welcher Suche fortgesetzt wird.
 */
  private void setzeDame( int spalte)
  {
// Loesung gefunden
    if( spalte > 7)
    { 
      DameLoesung loesung = new DameLoesung( spalten);
      loesungen.addElement( loesung);
      return;
    }

// keine Loesung gefunden, naechste Spalte durchsuchen
    for( int zeile = 0; zeile < 8; zeile++)
    {
      if( !istBedroht( zeile, spalte))
      {
// Setze Dame
        spalten[ spalte] = zeile;
        setzeBedrohung( zeile, spalte);

// Rekursion
        setzeDame( spalte + 1);

// Entferne Dame
        loescheBedrohung( zeile, spalte);
      }
    }
  }

/**
 * Ueberprueft Bedrohungen.
 * @param zeile Dame steht in Zeile zeile
 * @param spalte Dame steht in Spalte spalte
 */
  private boolean istBedroht( int zeile, int spalte)
  {
    return zeilen[ zeile] ||
           posDiag[ zeile + spalte] ||
           negDiag[ zeile - spalte + 7];
  }

/**
 * Setzt Bedrohungen.
 * @param zeile Dame steht in Zeile zeile
 * @param spalte Dame steht in Spalte spalte
 */
  private void setzeBedrohung( int zeile, int spalte)
  {
    zeilen[ zeile] = 
    posDiag[ zeile + spalte] =
    negDiag[ zeile - spalte + 7] = true;
  }

/**
 * Loescht Bedrohungen.
 * @param zeile Dame steht in Zeile zeile
 * @param spalte Dame steht in Spalte spalte
 */
  private void loescheBedrohung( int zeile, int spalte)
  {
    zeilen[ zeile] = 
    posDiag[ zeile + spalte] =
    negDiag[ zeile - spalte + 7] = false;
  }

/* ----------------------------------------------------- */
/**
 * Schachuebliche Darstellung aller Loesungen.
 * @return Loesungen
 */
  public String toString()
  {
    String str = "\n" + loesungen.anzahl()
               + " Loesungen:\n\n" + loesungen;
    return str;
  }
}
